#include "TamaP.h"
#include "math.h"

char * TamaClass_GetName(TamaClass tama_class)
{ return (tama_class->name); }

double TamaClass_GetProbability(TamaClass tama_class)
{ return (tama_class->probability); }

ObjList TamaClass_GetGCListList(TamaClass tama_class)
{ return (tama_class->gc_list_list); }

ObjList TamaClass_GetNextList(TamaClass tama_class)
{ return (tama_class->next_list); }

double TamaNext_GetPower(TamaNext tama_next)
{ return (tama_next->power); }

TamaClass Tamas_GetTamaClass(Tamas tamas)
{ return (tamas->tama_class); }

double * Tamas_GetX(Tamas tamas) { return (tamas->old_x); }
double * Tamas_GetY(Tamas tamas) { return (tamas->old_y); }
double * Tamas_GetZ(Tamas tamas) { return (tamas->old_z); }
double * Tamas_GetVx(Tamas tamas) { return (tamas->vx); }
double * Tamas_GetVy(Tamas tamas) { return (tamas->vy); }
double * Tamas_GetVz(Tamas tamas) { return (tamas->vz); }
int Tamas_GetNumber(Tamas tamas) { return (tamas->number); }

int TamaNext_GetNumber(TamaNext tama_next)
{ return (tama_next->number); }

TamaClass TamaNext_GetTamaClass(TamaNext tama_next)
{ return (tama_next->tama_class); }

/*===========================================================================*/
/* ƥ֥Ȥκ                                              */
/*===========================================================================*/

TamaClass TamaClass_Create(char * name,
			   int size,
			   double probability,
			   double air,
			   double gravity,
			   double transmission,
			   int after_image_length,
			   int fine,
			   ObjList gc_list_list,
			   ObjList next_list,
			   Disp disp)
{
  TamaClass tama_class;

  tama_class = (TamaClass)malloc(sizeof(_TamaClass));
  if (!tama_class) Error("TamaClass_Create", "Cannot allocate memory.");

  if ((name == NULL) || (name[0] == '\0')) name = "Unknown\0";
  tama_class->name = strdup(name);
  if (tama_class->name == NULL)
    Error("TamaClass_Create", "Cannot allocate memory.");
  tama_class->size = size;
  tama_class->probability = probability;
  tama_class->air = air;
  tama_class->gravity = gravity;
  tama_class->transmission = transmission;
  tama_class->after_image_length = after_image_length;
  tama_class->fine = fine;

  if (gc_list_list) tama_class->gc_list_list = gc_list_list;
  else              tama_class->gc_list_list = ObjList_Create();

  if (next_list) tama_class->next_list = next_list;
  else           tama_class->next_list = ObjList_Create();

  return (tama_class);
}

TamaNext TamaNext_Create(TamaClass tama_class, double power, int n)
{
  TamaNext next;

  next = (TamaNext)malloc(sizeof(_TamaNext));
  if (!next) Error("TamaNext_Create", "Cannot allocate memory.");

  next->tama_class = tama_class;
  next->power = power;
  next->number = n;

  return (next);
}

TamaClass TamaClass_Destroy(TamaClass tama_class)
{
  if (!tama_class) return (NULL);

  if (tama_class->name) free(tama_class->name);

  if (tama_class->gc_list_list)
    ObjList_Destroy(tama_class->gc_list_list);

  if (tama_class->next_list)
    ObjList_Destroy(tama_class->next_list);

  free(tama_class);

  return (NULL);
}

TamaNext TamaNext_Destroy(TamaNext next)
{
  if (!next) return (NULL);

  free(next);

  return (NULL);
}

/*===========================================================================*/
/* ǻѤؿ                                                        */
/*===========================================================================*/

/* TamaClass θ */

static TamaClass TamaClassList_SearchOrCreateTamaClass(ObjList list,
						       char * name,
						       Disp disp,
						       int default_size,
						       double default_air,
						       double default_gravity,
						       double default_transmission,
						       int default_after_image,
						       int fine)
{
  ObjListData current;
  TamaClass tama_class;

  if (!list) Error("TamaClassList_SearchOrCreateTamaClass",
		   "TamaClassList has not created.");

  for (current = ObjList_GetStart(list);
       !ObjList_IsEndEdge(list, current);
       current = ObjListData_GetNext(current)) {
    tama_class = (TamaClass)ObjListData_GetObj(current);
    if (!strcmp(tama_class->name, name)) {
      return (tama_class);
    }
  }

  /* ĤʤäˤϡꥹȤɲä */
  tama_class = TamaClass_Create(name,
				default_size,
				-1.0,
				default_air,
				default_gravity,
				default_transmission,
				default_after_image,
				fine,
				NULL, NULL, disp);
  ObjList_InsertObjToEnd(list, tama_class, (ObjDestructor)TamaClass_Destroy);

  return (tama_class);
}

/*---------------------------------------------------------------------------*/
/* ñμ                                                                */
/*---------------------------------------------------------------------------*/

static char * GetWord(Stream stream)
{
  char * word;
  word = Stream_GetWord(stream, NULL, 0,
			"",     /* ʸڤʬ */
			" \t",  /* ʸǤڤʬ */
			"!#$%", /* ȹԤƬʸ '#' ʤ */
			"\n",   /* ʸ '\n' ʤ */
			"\"\'", /* ʸ '\"' ʤ */
			"",     /* ʸƬɤߤȤФʸ */
			""      /* ʸɤߤȤФʸ */
			);
  return (word);
}

static char * GetWord2(Stream stream)
{
  char * word;
  word = Stream_GetWord(stream, NULL, 0,
			"",     /* ʸڤʬ */
			" \t",  /* ʸǤڤʬ */
			"",     /* ȹԤƬʸ '#' ʤ */
			"\n",   /* ʸ '\n' ʤ */
			"\"\'", /* ʸ '\"' ʤ */
			"",     /* ʸƬɤߤȤФʸ */
			""      /* ʸɤߤȤФʸ */
			);
  return (word);
}

static ObjList ReadNext(ObjList tama_class_list, char * word, Disp disp,
			int default_size,
			double default_air,
			double default_gravity,
			double default_transmission,
			int default_after_image,
			int fine)
{
  ObjList list;
  Stream stream;
  char * w1;
  char * w2;
  char * w3;
  TamaClass t;
  double power;
  int n;

  list = ObjList_Create();
  if (!word) return (list);
  stream = Stream_CreateFromCharacters(word);
  if (!stream) return (list);

  while ((w1 = GetWord2(stream)) != NULL) {
    t = TamaClassList_SearchOrCreateTamaClass(tama_class_list, w1, disp,
					      default_size, default_air,
					      default_gravity,
					      default_transmission,
					      default_after_image,
					      fine);
    w2 = GetWord2(stream);
    if ((w3 = GetWord2(stream)) != NULL) {
      power = atof(w2);
      n = atoi(w3);
      ObjList_InsertObjToEnd(list, TamaNext_Create(t, power, n),
			     (ObjDestructor)TamaNext_Destroy);
    }
    if (w1) free(w1);
    if (w2) free(w2);
    if (w3) free(w3);
  }

  Stream_Destroy(stream);

  return (list);
}

/* ǡɤ߹ */
ObjList TamaClassList_CreateFromStream(Disp disp,
				       ColorGCDatabase database,
				       Stream stream, char * background_color,
				       int fine)
{
  ObjList tama_class_list;
  TamaClass current = NULL;
  char * word;
  char * word2;
  Stream s;
  double default_probability = 0.1;
  int default_size = 3;
  double default_air = 0.05;
  double default_gravity = 0.3;
  double default_transmission = 60.0;
  int default_after_image = 10;

  tama_class_list = ObjList_Create();

  while ((word = GetWord(stream)) != NULL) {
    word2 = NULL;

    if (!strcmp(word, "Name")) {
      if ((word2 = GetWord(stream)) != NULL) {
	current =
	  TamaClassList_SearchOrCreateTamaClass(tama_class_list, word2, disp,
						default_size, default_air,
						default_gravity,
						default_transmission,
						default_after_image,
						fine);
      }
    } else if (!strcmp(word, "DefaultProbability")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_probability = atof(word2);
      }
    } else if (!strcmp(word, "DefaultSize")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_size = atoi(word2);
      }
    } else if (!strcmp(word, "DefaultAir")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_air = atof(word2);
      }
    } else if (!strcmp(word, "DefaultGravity")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_gravity = atof(word2);
      }
    } else if (!strcmp(word, "DefaultTransmission")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_transmission = atof(word2);
      }
    } else if (!strcmp(word, "DefaultAfterImage")) {
      if ((word2 = GetWord(stream)) != NULL) {
	default_after_image = atoi(word2);
      }
    } else {
      if (current) {
	if (!strcmp(word, "Size")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!strcmp(word2, "Default")) current->size = default_size;
	    else current->size = atoi(word2);
	  }
	} else if (!strcmp(word, "Probability")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if      (!strcmp(word2, "Default"))
	      current->probability = default_probability;
	    else if (!strcmp(word2, "None")) current->probability = -1.0;
	    else current->probability = atof(word2);
	  }
	} else if (!strcmp(word, "Air")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!strcmp(word2, "Default")) current->air = default_air;
	    else current->air = atof(word2);
	  }
	} else if (!strcmp(word, "Gravity")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!strcmp(word2, "Default")) current->gravity = default_gravity;
	    else current->gravity = atof(word2);
	  }
	} else if (!strcmp(word, "Transmission")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!strcmp(word2, "Default"))
	      current->transmission = default_transmission;
	    else current->transmission = atof(word2);
	  }
	} else if (!strcmp(word, "AfterImage")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (!strcmp(word2, "Default"))
	      current->after_image_length = default_after_image;
	    else current->after_image_length = atoi(word2);
	  }
	} else if (!strcmp(word, "Color")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (current->gc_list_list == NULL)
	      current->gc_list_list = ObjList_Create();
	    ObjList_Concatenate(current->gc_list_list,
				CreateColorGCListFromCharacters(database,
								word2,
								background_color,
								current->after_image_length,
								fine));

	  }
	} else if (!strcmp(word, "Next")) {
	  if ((word2 = GetWord(stream)) != NULL) {
	    if (current->next_list == NULL)
	      current->next_list = ObjList_Create();
	    if (strcmp(word2, "None"))
	      ObjList_Concatenate(current->next_list,
				  ReadNext(tama_class_list, word2, disp,
					   default_size,
					   default_air,
					   default_gravity,
					   default_transmission,
					   default_after_image,
					   fine));
	  }
	} else if (!strcmp(word, "End")) {
	  if (word)  free(word);
	  if (word2) free(word2);
	  return (tama_class_list);
	} else {
	  fprintf(stderr, "Unknown command : %s\n", word);
	}
      } else {
	fprintf(stderr,
		"Cannot specify class name. Command was ignored. : %s\n",
		word);
      }
    }
    if (word)  free(word);
    if (word2) free(word2);
  }

  return (tama_class_list);
}

/*===========================================================================*/
/* Tamas ֥Ȥ                                                  */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* Current 촹                                                        */
/*---------------------------------------------------------------------------*/

void Tamas_ChangeCurrent(Tamas tamas)
{
  double * p;

  p = tamas->old_x; tamas->old_x = tamas->new_x; tamas->new_x = p;
  p = tamas->old_y; tamas->old_y = tamas->new_y; tamas->new_y = p;
  p = tamas->old_z; tamas->old_z = tamas->new_z; tamas->new_z = p;
}

/*---------------------------------------------------------------------------*/
/* Tamas ֥Ȥΰư                                                  */
/*---------------------------------------------------------------------------*/

static ObjList Tamas_Move(Tamas tamas,
			  int x_min, int y_min, int x_max, int y_max,
			  int size)
{
  AfterImages after_images;
  int length;
  int i;
  double z;
  double step;
  ObjList gc_list;

  if (ObjList_IsEndEdge(tamas->tama_class->gc_list_list, tamas->gc_list))
    return (NULL);

  step = 100.0 / tamas->tama_class->fine;

  for (i = 0; i < tamas->number; i++) {

    tamas->vy[i] += tamas->tama_class->gravity * step;

    /* 񹳤η׻ */
    tamas->vx[i] *= (1.0 - tamas->tama_class->air * step);
    tamas->vy[i] *= (1.0 - tamas->tama_class->air * step);
    tamas->vz[i] *= (1.0 - tamas->tama_class->air * step);

    /* Ԥη׻ */
    tamas->new_z[i] = tamas->old_z[i] + tamas->vz[i] * step;
    z = 1.0 - tamas->new_z[i] * (1.0 / 1000.0);
    if (z < 0.01) z = 0.01;

    tamas->new_x[i] =
      tamas->old_x[i] + tamas->vx[i] * step * z * size * (1.0 / 1000.0);
    tamas->new_y[i] =
      tamas->old_y[i] + tamas->vy[i] * step * z * size * (1.0 / 1000.0);

    if (((tamas->new_x[i] < x_min) && (tamas->vx[i] < 0.0)) ||
	((tamas->new_x[i] > x_max) && (tamas->vx[i] > 0.0)) ||
	(tamas->new_y[i] > y_max)) {
    }
  }

  for (i = 0; i < tamas->number; i++) {
    if ( ((int)tamas->new_x[i] == (int)tamas->old_x[i]) &&
	 ((int)tamas->new_y[i] == (int)tamas->old_y[i]) )
      tamas->change_flags[i] = 0;
    else
      tamas->change_flags[i] = 1;
  }

  gc_list = ObjListData_GetObj(tamas->gc_list);

  return (gc_list);
}

AfterImages Tamas_Do(Tamas tamas, int x_min, int y_min, int x_max, int y_max,
		     int size)
{
  AfterImages after_images;
  ObjList gc_list;

  if (tamas == NULL) return (NULL);
  Tamas_ChangeCurrent(tamas);
  gc_list = Tamas_Move(tamas, x_min, y_min, x_max, y_max, size);
  if (gc_list == NULL) return (NULL);
  after_images = AfterImages_Create(tamas->tama_class->size, tamas->number,
				    tamas->old_x, tamas->old_y,
				    tamas->change_flags,
				    x_min, y_min, x_max, y_max,
				    ObjListData_GetObj(tamas->gc_list));
  tamas->gc_list = ObjListData_GetNext(tamas->gc_list);

  return (after_images);
}

/*---------------------------------------------------------------------------*/
/* Tamas ֥Ȥ                                                  */
/*---------------------------------------------------------------------------*/

Tamas Tamas_Create(TamaClass tama_class, double power, int number, int n,
		   double * x, double * y, double * z,
		   double * vx, double * vy, double * vz,
		   int x_min, int y_min, int x_max, int y_max)
{
  double rad, z_rad;
  Tamas tamas;
  double p;
  int i, j, a;

  tamas = (Tamas)malloc(sizeof(_Tamas));
  if (!tamas) Error("Tamas_Create", "Cannot allocate memory");

  tamas->tama_class = tama_class;

  tamas->change_flags = (int *)malloc(sizeof(int) * number * n);
  if (tamas->change_flags == NULL)
    Error("Tamas_Create", "Cannot allocate memory");

  tamas->x1 = (double *)malloc(sizeof(double) * number * n);
  if (tamas->x1 == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->y1 = (double *)malloc(sizeof(double) * number * n);
  if (tamas->y1 == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->z1 = (double *)malloc(sizeof(double) * number * n);
  if (tamas->z1 == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->x2 = (double *)malloc(sizeof(double) * number * n);
  if (tamas->x2 == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->y2 = (double *)malloc(sizeof(double) * number * n);
  if (tamas->y2 == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->z2 = (double *)malloc(sizeof(double) * number * n);
  if (tamas->z2 == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->vx = (double *)malloc(sizeof(double) * number * n);
  if (tamas->vx == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->vy = (double *)malloc(sizeof(double) * number * n);
  if (tamas->vy == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->vz = (double *)malloc(sizeof(double) * number * n);
  if (tamas->vz == NULL) Error("Tamas_Create", "Cannot allocate memory");

  tamas->old_x = tamas->x1;
  tamas->old_y = tamas->y1;
  tamas->old_z = tamas->z1;
  tamas->new_x = tamas->x2;
  tamas->new_y = tamas->y2;
  tamas->new_z = tamas->z2;

  tamas->gc_list = ObjList_GetStart(tamas->tama_class->gc_list_list);

  a = 0;
  for (i = 0; i < n; i++) {
    for (j = 0; j < number; j++) {
      z_rad = DoubleRand(M_PI * 2.0);
      p = cos(z_rad);
      rad = DoubleRand(M_PI * 2.0);
      tamas->change_flags[a] = 0;
      tamas->new_x[a] = x[i];
      tamas->new_y[a] = y[i];
      tamas->new_z[a] = z[i];
      tamas->vx[a] = cos(rad) * power * p
	+ vx[i] * tamas->tama_class->transmission * 0.01;
      tamas->vy[a] = sin(rad) * power * p
	+ vy[i] * tamas->tama_class->transmission * 0.01;
      tamas->vz[a] = power * sin(z_rad)
	+ vz[i] * tamas->tama_class->transmission * 0.01;
      if ( ((tamas->new_x[a] < x_min) && (tamas->vx[a] < 0.0)) ||
	   ((tamas->new_x[a] > x_max) && (tamas->vx[a] > 0.0)) ||
	   ((tamas->new_y[a] > y_max) && (tamas->vy[a] > 0.0)) ) {
	/* None */
      } else {
	a++;
      }
    }
  }
  tamas->number = a;

  return (tamas);
}

/*---------------------------------------------------------------------------*/
/* Tama ֥Ȥκ                                                   */
/*---------------------------------------------------------------------------*/

Tamas Tamas_Destroy(Tamas tamas)
{
  if (!tamas) return (NULL);

  if (tamas->change_flags) free(tamas->change_flags);
  if (tamas->x1) free(tamas->x1);
  if (tamas->y1) free(tamas->y1);
  if (tamas->z1) free(tamas->z1);
  if (tamas->x2) free(tamas->x2);
  if (tamas->y2) free(tamas->y2);
  if (tamas->z2) free(tamas->z2);
  if (tamas->vx) free(tamas->vx);
  if (tamas->vy) free(tamas->vy);
  if (tamas->vz) free(tamas->vz);

  free(tamas);

  return (NULL);
}

