#include "XHanabiP.h"
#include "xhanabi_conf.h"

static Stream Stream_CreateFromFileOrCharacters(char * dirname,
						char * filename,
						char * default_xhanabi_conf)
{
  Stream stream;
  char * work;

  stream = Stream_CreateFromFile(filename);
  if (stream != NULL) return (stream);

  /* dirname  "" äȤˡ./[filename] Τ褦ʥե */
  /* Ĵ٤Τǡ1 ǤϤʤ 3 äƤ롥                     */
  work =
    (char *)malloc(sizeof(char) * (strlen(dirname) + strlen(filename) + 3));

  strcpy(work, "./");
  strcat(work, filename);

  stream = Stream_CreateFromFile(filename);
  if (stream != NULL) goto stream_return;

  strcpy(work, dirname);
  strcat(work, "/");
  strcat(work, filename);

  stream = Stream_CreateFromFile(filename);
  if (stream != NULL) goto stream_return;

  stream = Stream_CreateFromCharacters(default_xhanabi_conf);

  stream_return :

  free(work);
  return (stream);
}

XHanabi XHanabi_Create(char * display_name, char * background_color,
		       char * dirname, char * filename,
		       int studying_flag, int cache_flag, int cache_size,
		       int hash_number, int fine)
{
  XHanabi xhanabi;
  Stream stream;

  xhanabi = (XHanabi)malloc(sizeof(_XHanabi));
  if (!xhanabi) Error("XHanabi_Create", "Cannot allocate mamory");

  /*  */
  InitializeRand();

  xhanabi->disp = Disp_Create(display_name, background_color);

  xhanabi->color_gc_database = ColorGCDatabase_Create(xhanabi->disp,
						      studying_flag,
						      cache_flag,
						      cache_size,
						      hash_number,
						      background_color);

  /* TamaClass ɤ߹ */
  stream = 
    Stream_CreateFromFileOrCharacters(dirname, filename, default_xhanabi_conf);

  xhanabi->tama_class_list =
    TamaClassList_CreateFromStream(xhanabi->disp, xhanabi->color_gc_database,
				   stream, background_color, fine);
  Stream_Destroy(stream);

  xhanabi->tamas_list = ObjList_Create();
  xhanabi->after_images_list = ObjList_Create();

  return (xhanabi);
}

int XHanabi_Destroy(XHanabi xhanabi)
{
  if (!xhanabi) return (0);
  if (xhanabi->tama_class_list) ObjList_Destroy(xhanabi->tama_class_list);
  if (xhanabi->tamas_list) ObjList_Destroy(xhanabi->tamas_list);
  if (xhanabi->after_images_list) ObjList_Destroy(xhanabi->after_images_list);
  if (xhanabi->color_gc_database) ColorGCDatabase_Destroy(xhanabi->color_gc_database);
  if (xhanabi->disp) Disp_Destroy(xhanabi->disp);
  free(xhanabi);
  return (0);
}

static int XHanabi_DrawAllTamaClasses(XHanabi xhanabi)
{
  TamaClass tama_class;
  ObjList gc_list_list;
  ObjList gc_list;
  ObjListData a, b, c;
  GC gc;

  int x, y;

  x = 10;
  y = 10;

  for (a = ObjList_GetStart(xhanabi->tama_class_list);
       !ObjList_IsEndEdge(xhanabi->tama_class_list, a);
       a = ObjListData_GetNext(a)) {
    tama_class = (TamaClass)ObjListData_GetObj(a);
    gc_list_list = TamaClass_GetGCListList(tama_class);

    for (b = ObjList_GetStart(gc_list_list);
	 !ObjList_IsEndEdge(gc_list_list, b);
	 b = ObjListData_GetNext(b)) {

      gc_list = ObjListData_GetObj(b);

      x = 10;
      for (c = ObjList_GetStart(gc_list);
	   !ObjList_IsEndEdge(gc_list, c);
	   c = ObjListData_GetNext(c)) {
	gc = ColorGC_GetGC(ObjListData_GetObj(c));
	Disp_DrawFilledCircle(xhanabi->disp, gc, x, y, 5);
	Disp_Flush(xhanabi->disp);
	x += 10;
      }
      y += 10;
    }
  }
  return (0);
}

int XHanabi_Start(XHanabi xhanabi, int tama_max_number, int wait_time)
{
  TamaClass tama_class;
  Tamas tamas;
  TamaNext tama_next;
  ObjList next_list;
  ObjListData current;
  ObjListData current2;
  AfterImages after_images;
  int i;
  double d;
  double x, y, z, vx, vy, vz;
  GC gc;
  /*
  XHanabi_DrawAllTamaClasses(xhanabi);
  while(1){}
  */

  while (1) {

    /*  Tamas  */
    if (ObjList_GetLength(xhanabi->tamas_list) < tama_max_number) {

      for (current = ObjList_GetStart(xhanabi->tama_class_list);
	   !ObjList_IsEndEdge(xhanabi->tama_class_list, current);
	   current = ObjListData_GetNext(current)) {
	tama_class = (TamaClass)ObjListData_GetObj(current);
	if (TamaClass_GetProbability(tama_class) > 0.0) {
	  if (TamaClass_GetProbability(tama_class) > DoubleRand(100.0)) {
	    x = (double)Rand(Disp_GetWidth( xhanabi->disp));
	    y = (double)Rand(Disp_GetHeight(xhanabi->disp));
	    z = DoubleRand(400.0) - 200.0;
	    vx = 0.0;
	    vy = 0.0;
	    vz = 0.0;
	    y -= Disp_GetHeight(xhanabi->disp) / 4;

	    tamas = Tamas_Create(tama_class, 0.0, 1, 1,
				 &x, &y, &z, &vx, &vy, &vz,
				 0, 0,
				 Disp_GetWidth( xhanabi->disp) - 1,
				 Disp_GetHeight(xhanabi->disp) - 1);
	    ObjList_InsertObjToEnd(xhanabi->tamas_list,
				   tamas,
				   (ObjDestructor)Tamas_Destroy);

	  }
	}
      }
    }

    /* Tamas ν */
    for (current = ObjList_GetStart(xhanabi->tamas_list);
	 !ObjList_IsEndEdge(xhanabi->tamas_list, current);
	 current = ObjListData_GetNext(current)) {

      tamas = (Tamas)ObjListData_GetObj(current);
      tama_class = Tamas_GetTamaClass(tamas);

      after_images = Tamas_Do(tamas,
			      0, 0, 
			      Disp_GetWidth( xhanabi->disp) - 1,
			      Disp_GetHeight(xhanabi->disp) - 1,
			      Disp_GetWidth( xhanabi->disp));

      if (after_images == NULL) { /* ǳԤ */
	next_list = TamaClass_GetNextList(tama_class);
	if (next_list) {
	  for (current2 = ObjList_GetStart(next_list);
	       !ObjList_IsEndEdge(next_list, current2);
	       current2 = ObjListData_GetNext(current2)) {
	    tama_next = (TamaNext)ObjListData_GetObj(current2);

	    ObjList_InsertObjToEnd(xhanabi->tamas_list,
				   Tamas_Create(TamaNext_GetTamaClass(tama_next),
						TamaNext_GetPower(tama_next),
						TamaNext_GetNumber(tama_next),
						Tamas_GetNumber(tamas),
						Tamas_GetX(tamas),
						Tamas_GetY(tamas),
						Tamas_GetZ(tamas),
						Tamas_GetVx(tamas),
						Tamas_GetVy(tamas),
						Tamas_GetVz(tamas),
						0, 0, 
						Disp_GetWidth( xhanabi->disp) - 1,
						Disp_GetHeight(xhanabi->disp) - 1),
				   (ObjDestructor)Tamas_Destroy);
	  }
	}
	current = ObjList_DeleteObjToPrev(xhanabi->tamas_list, current);
      } else {
	ObjList_InsertObjToEnd(xhanabi->after_images_list, after_images,
			       (ObjDestructor)AfterImages_Destroy);
      }
    }

    /* ν */
    for (current = ObjList_GetStart(xhanabi->after_images_list);
	 !ObjList_IsEndEdge(xhanabi->after_images_list, current);
	 current = ObjListData_GetNext(current)) {
      after_images = (AfterImages)ObjListData_GetObj(current);

      gc = AfterImages_GetGC(after_images);

      if (gc == NULL) {
	current = ObjList_DeleteObjToPrev(xhanabi->after_images_list, current);
      } else {
	Disp_DrawFilledCircles(xhanabi->disp, gc,
			       AfterImages_GetArcs(after_images),
			       AfterImages_GetNumber(after_images));
	AfterImages_MoveNext(after_images);
	/*usleep(10);*/
	/* X ФϢ³٤ͿʤΥ */
      }
    }

    Disp_Flush(xhanabi->disp);
    usleep(100 + wait_time);
  }

  return (0);
}

/*****************************************************************************/
/* End of File.                                                              */
/*****************************************************************************/
